--imports
local assert = require("justo.assert")
local justo = require("justo")
local suite, test, init = justo.suite, justo.test, justo.init
local Trans = require("dogma.trans.js.Trans")
local Parser = require("dogma.syn.Parser")

--Suite.
return suite("dogma.trans.js._.ExpTrans", function()
  local trans, parser

  ----------
  -- init --
  ----------
  init("*", function()
    parser = Parser.new()
    trans = Trans.new()
    trans:transform(parser)
  end):title("Create transformer")

  --------------
  -- terminal --
  --------------
  suite("terminal", function()
    test("nop", function()
      parser:parse("nop")
      assert(trans:next()):eq("dogma.nop();\n")
    end):tags("nop")

    suite("name", function()
      test("name", function()
        parser:parse("x")
        assert(trans:next()):eq("x;\n")
      end)

      test("keyword", function()
        parser:parse("class")
        assert(trans:next()):eq("class_;\n")
      end)
    end)

    test("regular expression", function()
      parser:parse('`pattern"`')
      assert(trans:next()):eq('re("pattern\"");\n')
    end)

    test("123", function()
      parser:parse("123")
      assert(trans:next()):eq("123;\n")
    end)

    test("true", function()
      parser:parse("true")
      assert(trans:next()):eq("true;\n")
    end)

    test("false", function()
      parser:parse("false")
      assert(trans:next()):eq("false;\n")
    end)

    test("nil", function()
      parser:parse("nil")
      assert(trans:next()):eq("null;\n")
    end)

    suite("text", function()
      test('"text"', function()
        parser:parse('"text"')
        assert(trans:next()):eq('"text";\n')
      end)

      test('"text" - with new lines', function()
        parser:parse('"text1\ntext2"')
        assert(trans:next()):eq('"text1\\ntext2";\n')
      end)

      test('$"text"', function()
        parser:parse('$"text"')
        assert(trans:next()):eq('`text`;\n')
      end)
    end)

    suite("json", function()
      test("json", function()
        parser:parse("json")
        assert(trans:next()):eq("json;\n")
      end)

      test("json.encode", function()
        parser:parse("json.encode")
        assert(trans:next()):eq("json.encode;\n")
      end)

      test("json.decode", function()
        parser:parse("json.decode")
        assert(trans:next()):eq("json.decode;\n")
      end)

      test("json.decode(serialized)", function()
        parser:parse("json.decode(serialized)")
        assert(trans:next()):eq("json.decode(serialized);\n")
      end)

      test("json.decode<MyStruct>(serialized)", function()
        parser:parse("json.decode<MyStruct>(serialized)")
        assert(trans:next()):eq("json.decode(serialized, MyStruct);\n")
      end)
    end)

    test("self", function()
      parser:parse("self")
      assert(trans:next()):eq("this;\n")
    end)

    test("(1+2)", function()
      parser:parse("(1+2)")
      assert(trans:next()):eq("(1+2);\n")
    end)

    test("()", function()
      parser:parse("(\n)")
      assert(trans:next()):eq("null;\n")
    end)

    test("(exp, exp)", function()
      parser:parse("(\n1+2\n3+4\n)")
      assert(trans:next()):eq("((1+2), (3+4));\n")
    end)

    test("super", function()
      parser:parse("super")
      assert(trans:next()):eq("super;\n")
    end)

    suite("[for]", function()
      test("[for x in y do z]", function()
        parser:parse("l = [for i in mylist do i*2]")
        assert(trans:next()):eq("(l=(mylist).map((i)=>(i*2)));\n")
      end)

      test("[for a, b in c do d]", function()
        parser:parse("l = [for a, b in myobj do a*b]")
        assert(trans:next()):eq("(l=(entries(myobj)).map(([a, b])=>(a*b)));\n")
      end)
    end)

    suite("list", function()
      test("[]", function()
        parser:parse("a + [] + b")
        assert(trans:next()):eq("((a+[])+b);\n")
      end)

      test("[Exp]", function()
        parser:parse("a + [123] + b")
        assert(trans:next()):eq("((a+[123])+b);\n")
      end)

      test("[Exp, Exp]", function()
        parser:parse("a + [123, 456] + b")
        assert(trans:next()):eq("((a+[123, 456])+b);\n")
      end)
    end)

    suite("set", function()
      test("[||]", function()
        parser:parse("a + [||] + b")
        assert(trans:next()):eq("((a+dogma.set())+b);\n")
      end)

      test("[|Exp|]", function()
        parser:parse("a + [|123|] + b")
        assert(trans:next()):eq("((a+dogma.set(123))+b);\n")
      end)

      test("[|Exp, Exp|]", function()
        parser:parse("a + [|123, 456|] + b")
        assert(trans:next()):eq("((a+dogma.set(123, 456))+b);\n")
      end)
    end)

    suite("map", function()
      test("{}", function()
        parser:parse("a + {}")
        assert(trans:next()):eq("(a+{});\n")
      end)

      test("{Name = Exp}", function()
        parser:parse("a + {x = 1+2}")
        assert(trans:next()):eq('(a+{["x"]: (1+2)});\n')
      end)

      test("{Name = Exp, Name, {Name} = Exp, [Name] = Exp, ...variable}", function()
        parser:parse("a + {x = 1+2, y, {z} = obj, [x+y] = obj2, ...variable}")
        assert(trans:next()):eq('(a+{["x"]: (1+2), ["y"]: y, ["z"]: obj.z, [(x+y)]: obj2, ...variable});\n')
      end)

      test("with optional fields", function()
        parser:parse("x = {a = 11, if cond1 then b = 22, if cond2 then c, d}")
        assert(trans:next()):eq(
          '(x=Object.assign({}, {["a"]: 11, ["d"]: d}, (cond1 ? {["b"]: 22} : {}), (cond2 ? {["c"]: c} : {})));\n'
        )
      end)
    end):tags("map")

    suite("xml", function()
      test("<element prop=(value) prop=value prop prop=(value) {prop}=(value) ...obj ...(exp) />", function()
        parser:parse("<element prop1=(val1) prop2=val2 prop3 prop4=(val4) {prop5}=(val5) ...obj ...(exp) />", nil, {react=true})
        assert(trans:next()):eq('(<element prop1={val1} prop2={val2} prop3={prop3} prop4={val4} prop5={val5.prop5} {...obj} {...exp}></element>);\n')
      end)

      test("check literals as values", function()
        parser:parse('<ele.sub map={x=1, y=2} auto list=[1,2,3] num=123 text="bonjour" bool=true func=fn() = 123 end />', nil, {react=true})
        assert(trans:next()):eq('(<ele.sub map={{["x"]: 1, ["y"]: 2}} auto={auto} list={[1, 2, 3]} num={123} text={"bonjour"} bool={true} func={(() => { {return 123;} })}></ele.sub>);\n')
      end)

      test("content is expression", function()
        parser:parse('<div>1+2+3</div>', nil, {react=true})
        assert(trans:next()):eq('(<div>{((1+2)+3)}</div>);\n')
      end)

      test("content is multicontent", function()
        parser:parse('<div><one/>(children)<two/></div>', nil, {react=true})
        assert(trans:next()):eq('(<div><one></one>{children}<two></two></div>);\n')
      end)
    end):tags("xml")

    suite("proc", function()
      test("proc() x + y end", function()
        parser:parse("a + proc() x + 1 end")
        assert(trans:next()):eq("(a+(() => { {(x+1);} }));\n")
      end)
    end)

    suite("fn", function()
      test("x = fn(x: *map) = x end", function()
        parser:parse("x = fn(x: *map) = x end")
        assert(trans:next()):eq('(x=((x) => { /* istanbul ignore next */ dogma.expect("x", x, map);x = dogma.copy(x);{return x;} }));\n')
      end)

      test("fn() end", function()
        parser:parse("a + fn() end")
        assert(trans:next()):eq("(a+(() => { {} }));\n")
      end)

      test("fn(done) => done(_) ... end", function()
        parser:parse("a + fn(done) => done(_)\nx+y\nend")
        assert(trans:next()):eq("(a+((done) => { try { /* istanbul ignore next */ dogma.expect(\"done\", done);{(x+y);} } catch (_) { done(_) }}));\n")
      end)

      test("fn() { ... }", function()
        parser:parse("a + fn() { #[!insert /* ignore */] if x then y else z }")
        assert(trans:next()):eq("(a+(() => { {/* ignore */if (x) {y;} else {z;}} }));\n")
      end)

      test("fn(Name) end", function()
        parser:parse("a + fn(x) end")
        assert(trans:next()):eq('(a+((x) => { /* istanbul ignore next */ dogma.expect("x", x);{} }));\n')
      end)

      test("fn(Name) -> Name end", function()
        parser:parse("a + fn(x) -> x end")
        assert(trans:next()):eq('(a+((x) => { /* istanbul ignore next */ dogma.expect("x", x);{} return x; }));\n')
      end)

      test("fn(Name=>[x,y]) = x + y end", function()
        parser:parse("a + fn(params=>[x,y]) = x+y end")
        assert(trans:next()):eq('(a+((params) => { /* istanbul ignore next */ dogma.expect("params", params, list);let [x,y] = params;{return (x+y);} }));\n')
      end)

      test("fn(Name=>{x,y}) = x + y end", function()
        parser:parse("a + fn(params=>{x,y}) = x+y end")
        assert(trans:next()):eq('(a+((params) => { /* istanbul ignore next */ dogma.expect("params", params, map);let {x,y} = params;{return (x+y);} }));\n')
      end)

      test("fn(Name?) end", function()
        parser:parse("a + fn(x?) end")
        assert(trans:next()):eq("(a+((x) => { {} }));\n")
      end)

      test("fn(Name : num) end", function()
        parser:parse("a + fn(x:num) end")
        assert(trans:next()):eq('(a+((x) => { /* istanbul ignore next */ dogma.expect("x", x, num);{} }));\n')
      end)

      test("fn(Name ? : num) end", function()
        parser:parse("a + fn(x?:num) end")
        assert(trans:next()):eq('(a+((x) => { /* istanbul ignore next */ if (x != null) dogma.expect("x", x, num);{} }));\n')
      end)

      test("fn(Name, Name) = Exp end", function()
        parser:parse("a + fn(x, y) = x - y end")
        assert(trans:next()):eq('(a+((x, y) => { /* istanbul ignore next */ dogma.expect("x", x);/* istanbul ignore next */ dogma.expect("y", y);{return (x-y);} }));\n')
      end)

      test("fn(Name:{fld:[type]}) = Exp end", function()
        parser:parse("a = fn(params:{src:text}) = params.src.join() end")
        assert(trans:next()):eq('(a=((params) => { /* istanbul ignore next */ dogma.expect(\"params\", params, dogma.intf("inline", {src: {optional: false, type: text}}));{return params.src.join();} }));\n')
      end)
    end)

    test("native(literal)", function()
      parser:parse('x = native("new Proxy(target, handler)")')
      assert(trans:next()):eq("(x=new Proxy(target, handler));\n")
    end):tags("native")

    test("const(literal)", function()
      parser:parse('x = const(obj)')
      assert(trans:next()):eq("(x=dogma.immutable(obj));\n")
    end):tags("const")

    test("copy(literal)", function()
      parser:parse('x = copy(obj)')
      assert(trans:next()):eq("(x=dogma.copy(obj));\n")
    end):tags("copy")

    suite("overload()", function()
      test("overload()", function()
        parser:parse('overload()')
        assert(trans:next()):eq("dogma.overload(Array.from(arguments), []);\n")
      end)

      test("overload(name, ?name, ...)", function()
        parser:parse('overload(name, ?name, ...)')
        assert(trans:next()):eq("dogma.overload(Array.from(arguments), [dogma.OverloadParam({kind: 'm', type: name}), dogma.OverloadParam({kind: 'o', type: name}), dogma.OverloadParam({kind: '...'})]);\n")
      end)

      test("overload(name[])", function()
        parser:parse('overload(name[])')
        assert(trans:next()):eq("dogma.overload(Array.from(arguments), [dogma.OverloadParam({kind: 'm', type: dogma.TypeDef({name: 'inline', types: [name], min: 0, max: null})})]);\n")
      end)
    end):tags("overload")

    suite("await()", function()
      test("await(Exp)", function()
        parser:parse("x = await(1+2)")
        assert(trans:next()):eq("(x=(0,await((1+2))));\n")
      end)

      test("await(Exp, Exp)", function()
        parser:parse("x = await(1+2, 3+4)")
        assert(trans:next()):eq("(x=(0,await(Promise.all([(1+2), (3+4)]))));\n")
      end)

      test("await(...Exp)", function()
        parser:parse("x = await(...array)")
        assert(trans:next()):eq("(x=(0,await(Promise.all(array))));\n")
      end)
    end):tags("await")

    suite("pawait()", function()
      test("pawait(Exp)", function()
        parser:parse("l = pawait(returnOnePromise())")
        assert(trans:next()):eq("(l=await(dogma.pawait(() => returnOnePromise())));\n")
      end):tags("pawait")

      test("pawait(Exp, Exp)", function()
        parser:parse("x = pawait(1+2, 3+4)")
        assert(trans:next()):like("^let $$aux%d+;%(x=%($$aux%d+ = %(0,await%(Promise%.allSettled%(%[%(1%+2%), %(3%+4%)%]%)%)%).map%(%(i%) => %[i.status == 'fulfilled', i.value || i.reason%]%), %[!$$aux%d+.find%(%(i%) => i%[0%] === false%), $$aux%d+%]%)%);\n")
      end)

      test("pawait(...Exp)", function()
        parser:parse("x = pawait(...array)")
        assert(trans:next()):like("^let $$aux%d+;%(x=%($$aux%d+ = %(0,await%(Promise%.allSettled%(array%)%)%).map%(%(i%) => %[i.status == 'fulfilled', i.value || i.reason%]%), %[!$$aux%d+.find%(%(i%) => i%[0%] === false%), $$aux%d+%]%)%);\n")
      end)
    end):tags("pawait")

    test("cwait(Exp)", function()
      parser:parse("x = cwait(myfn(done))")
      assert(trans:next()):eq("(x=(0,await(new Promise((rslv, rej) => { function done(err, rslt) { if (err) rej(err); else rslv(rslt); } myfn(done); }))));\n")
    end):tags("cwait")

    test("pcwait(Exp)", function()
      parser:parse("l = pcwait(callable(done))")
      assert(trans:next()):eq("(l=(0,await(new Promise((rslv) => { function done(err, rslt) { if (err) rslv([false, err]); else rslv([true, rslt]); } callable(done); }))));\n")
    end):tags("pcwait")

    test("pwait(Exp)", function()
      parser:parse("l = pwait(setTimeout(done, 750, 750))")
      assert(trans:next()):eq("(l=dogma.pwait((done) => {setTimeout(done, 750, 750);}));\n")
    end):tags("pwait")

    test("use(Exp)", function()
      parser:parse('l = use("path")')
      assert(trans:next()):eq('(l=dogma.use(require("path")));\n')
    end):tags("use")

    test("peval(Exp)", function()
      parser:parse("x = peval(1+2)")
      assert(trans:next()):eq("(x=dogma.peval(() => {return (1+2);}));\n")
    end)

    test("catch(Exp)", function()
      parser:parse("x = catch(1+2)")
      assert(trans:next()):eq("(x=dogma.catchError(() => (1+2)));\n")
    end)

    suite("emit()", function()
      test("emit(event, arg)", function()
        parser:parse('emit("evento", 123)')
        assert(trans:next()):eq('dogma.emit(this, "evento", 123);\n')
      end)

      test("emit(event, literal map)", function()
        parser:parse('emit("evento", x=, y=1, z)')
        assert(trans:next()):eq('dogma.emit(this, "evento", {x: x, y: 1, z: z});\n')
      end)
    end):tags("emit")

    suite("expect()", function()
      test("expect(x)", function()
        parser:parse("expect(x)")
        assert(trans:next()):eq("dogma.expect('x', x);\n")
      end)

      test("expect(x.y)", function()
        parser:parse("expect(x.y)")
        assert(trans:next()):eq("dogma.expect('x.y', x.y);\n")
      end)

      test("expect(x, types)", function()
        parser:parse("expect(x, [text, num])")
        assert(trans:next()):eq("dogma.expect('x', x, [text, num]);\n")
      end)
    end)

    suite("throw", function()
      test("throw(Exp)", function()
        parser:parse([[throw("my error.")]])
        assert(trans:next()):eq('dogma.raise("my error.");\n')
      end)

      test("throw(Exp, Exp)", function()
        parser:parse([[throw("%s", "my error.")]])
        assert(trans:next()):eq('dogma.raise("%s", "my error.");\n')
      end)

      test("throw(Exp, Exp, Exp)", function()
        parser:parse([[throw("%s: %s", "internal error", "my message.")]])
        assert(trans:next()):eq('dogma.raise("%s: %s", "internal error", "my message.");\n')
      end)
    end):tags("throw")

    suite("if..then..else..end", function()
      test("if..then..else..end", function()
        parser:parse([[(if x then x+1 else x+2 end) + 3]])
        assert(trans:next()):eq("((x ? (x+1) : (x+2))+3);\n")
      end)

      test("if..then..end", function()
        parser:parse([[(if x then x+1 end) + 3]])
        assert(trans:next()):eq("((x ? (x+1) : null)+3);\n")
      end)
    end)

    suite("with..do", function()
      test("without else", function()
        parser:parse("(with x do\nif 1 then 2\nif 3 then 4\nend)")
        assert(trans:next()):like("^let $$aux%d+;%(%($$aux%d+ = x%),%(%($$aux%d+ == 1%) %? %(2%) : %($$aux%d+ == 3%) %? %(4%) :  null%)%);\n")
      end)

      test("with else", function()
        parser:parse("(with x do\nif 1 then 2\nif 3 then 4\nelse 5\nend)")
        assert(trans:next()):like("^let $$aux%d+;%(%($$aux%d+ = x%),%(%($$aux%d+ == 1%) %? %(2%) : %($$aux%d+ == 3%) %? %(4%) : 5%)%);\n")
      end)
    end):tags("with-do")
  end)

  ------------------
  -- non-terminal --
  ------------------
  suite("non-terminal", function()
    --------------
    -- unary op --
    --------------
    suite("unary op", function()
      test("<<<", function()
        parser:parse("<<<args")
        assert(trans:next()):eq("dogma.lshift(args);\n")
      end)

      test(">>>", function()
        parser:parse(">>>args")
        assert(trans:next()):eq("dogma.rshift(args);\n")
      end)

      test("...", function()
        parser:parse("...args")
        assert(trans:next()):eq("...(args);\n")
      end)

      test(". Name", function()
        parser:parse(".x")
        assert(trans:next()):eq("this.x;\n")
      end)

      test(": Name", function()
        parser:parse(":x")
        assert(trans:next()):eq("this._x;\n")
      end)

      test("! Name", function()
        parser:parse("!x")
        assert(trans:next()):eq("this._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_x;\n")
      end)

      test("$ Name", function()
        parser:parse("$x")
        assert(trans:next()):eq("ps.env.x;\n")
      end)

      test("not x", function()
        parser:parse("not x")
        assert(trans:next()):eq("!(x);\n")
      end)

      test("~x", function()
        parser:parse("~x")
        assert(trans:next()):eq("~(x);\n")
      end)

      test("+x", function()
        parser:parse("+x")
        assert(trans:next()):eq("+(x);\n")
      end)

      test("-x", function()
        parser:parse("-x")
        assert(trans:next()):eq("-(x);\n")
      end)
    end):tags("unaryop")

    ------------
    -- bin op --
    ------------
    suite("bin op", function()
      test("op", function(params)
        parser:parse("x " .. params[1] .. " y")
        assert(trans:next()):eq(("(x" .. params[1] .. "y);\n"))
      end):iter(
        {subtitle = "+", params = "+"},
        {subtitle = "-", params = "-"},
        {subtitle = "*", params = "*"},
        {subtitle = "**", params = "**"},
        {subtitle = "/", params = "/"},
        {subtitle = "%", params = "%"},
        {subtitle = "==", params = "=="},
        {subtitle = "!=", params = "!="},
        {subtitle = "===", params = "==="},
        {subtitle = "!==", params = "!=="},
        {subtitle = "<", params = "<"},
        {subtitle = "<=", params = "<="},
        {subtitle = ">", params = ">"},
        {subtitle = ">=", params = ">="},
        {subtitle = "=", params = "="},
        {subtitle = "+=", params = "+="},
        {subtitle = "-=", params = "-="},
        {subtitle = "*=", params = "*="},
        {subtitle = "**=", params = "**="},
        {subtitle = "/=", params = "/="},
        {subtitle = "%=", params = "%="},
        {subtitle = ">>=", params = ">>="},
        {subtitle = "<<=", params = "<<="},
        {subtitle = "|=", params = "|="},
        {subtitle = "&=", params = "&="},
        {subtitle = "^=", params = "^="}
      )

      test("&&", function()
        parser:parse("x+y && a+b")
        assert(trans:next()):eq("dogma.pevalAnd(() => {(x+y)}, () => {return (a+b);});\n")
      end)

      test("||", function()
        parser:parse("x+y || a+b")
        assert(trans:next()):eq("dogma.pevalOr(() => {return (x+y);}, () => {return (a+b);});\n")
      end)

      suite("??", function()
        test("once", function()
          parser:parse("x ?? y")
          assert(trans:next()):eq("(x ?? y);\n")
        end)

        test("twice", function()
          parser:parse("x ?? y ?? z")
          assert(trans:next()):eq("((x ?? y) ?? z);\n")
        end)
      end)

      test("=~", function()
        parser:parse("status =~ FAILED")
        assert(trans:next()):eq('(status = dogma.enumGet(status, "FAILED"));\n')
      end)

      test("==~", function()
        parser:parse("status ==~ FAILED")
        assert(trans:next()):eq('dogma.enumEq(status, "FAILED");\n')
      end)

      test("!=~", function()
        parser:parse("status !=~ FAILED")
        assert(trans:next()):eq('(!dogma.enumEq(status, "FAILED"));\n')
      end)

      test("==^", function()
        parser:parse("status ==^ FAILED")
        assert(trans:next()):eq('dogma.deepEq(status, "FAILED");\n')
      end)

      test("!=^", function()
        parser:parse("status !=^ FAILED")
        assert(trans:next()):eq('(!dogma.deepEq(status, "FAILED"));\n')
      end)

      suite("?=", function()
        test("Name ?= Exp", function()
          parser:parse("x ?= 123")
          assert(trans:next()):eq("(x = x ?? 123);\n")
        end)

        test("Name.Name ?= Exp", function()
          parser:parse("x.y ?= 123")
          assert(trans:next()):eq("(x.y = x.y ?? 123);\n")
        end)

        test("Name:Name ?= Exp", function()
          parser:parse("x:y ?= 123")
          assert(trans:next()):eq("(x._y = x._y ?? 123);\n")
        end)

        test("Name[Ix] ?= Exp", function()
          parser:parse("x[y] ?= 123")
          assert(trans:next()):eq('dogma.setItem("=", x, y, dogma.getItem(x, y) ?? 123);\n')
        end)
      end)

      suite(":=", function()
        test(".Name := Exp", function()
          parser:parse(".x := 123")
          assert(trans:next()):eq('Object.defineProperty(this, "x", {value: 123, enum: true});\n')
        end)

        test(":Name := Exp", function()
          parser:parse(":x := 123")
          assert(trans:next()):eq('Object.defineProperty(this, "_x", {value: 123});\n')
        end)

        test("!Name := Exp", function()
          parser:parse("!x := 123")
          assert(trans:next()):eq('Object.defineProperty(this, "_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_x", {value: 123});\n')
        end)

        test("Name . Name := Exp", function()
          parser:parse("x.y := 123")
          assert(trans:next()):eq('Object.defineProperty(x, "y", {value: 123, enum: true});\n')
        end)

        test("Name : Name := Exp", function()
          parser:parse("x:y := 123")
          assert(trans:next()):eq('Object.defineProperty(x, "_y", {value: 123});\n')
        end)

        test("Name ! Name := Exp", function()
          parser:parse("x!y := 123")
          assert(trans:next()):eq('Object.defineProperty(x, "_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_y", {value: 123});\n')
        end)

        test("x[y]:=z", function()
          parser:parse("x[y]:=z")
          assert(trans:next()):eq('dogma.setItem("=", x, y, z);\n')
        end)
      end)

      suite(".=", function()
        test(":Name .= Exp", function()
          parser:parse(":x .= 123")
          assert(trans:next()):eq('Object.defineProperty(this, "_x", {value: 123, writable: true});Object.defineProperty(this, "x", {enum: true, get() { return this._x; }});\n')
        end)
      end)

      test(">>>", function()
        parser:parse("x >>> y")
        assert(trans:next()):eq("dogma.rshift(x, y);\n")
      end)

      test("<<<", function()
        parser:parse("x <<< y")
        assert(trans:next()):eq("dogma.lshift(x, y);\n")
      end)

      test("and", function()
        parser:parse("x and y")
        assert(trans:next()):eq("(x&&y);\n")
      end)

      test("or", function()
        parser:parse("x or y")
        assert(trans:next()):eq("((x) || (y));\n")
      end)

      suite(".", function()
        test("read", function()
          parser:parse("x.y")
          assert(trans:next()):eq("x.y;\n")
        end)

        test("super.method()", function()
          parser:parse("super.m()")
          assert(trans:next()):eq('dogma.super(this, "m")();\n')
        end)

        test("write", function()
          parser:parse("x.y=z")
          assert(trans:next()):eq("(x.y=z);\n")
        end)
      end)

      test("?", function()
        parser:parse("x?y")
        assert(trans:next()):eq("(x != null ? x.y : null);\n")
      end)

      suite(":", function()
        test("super:method()", function()
          parser:parse("super:m()")
          assert(trans:next()):eq('dogma.super(this, "_m")();\n')
        end)

        test("x:y", function()
          parser:parse("x:y")
          assert(trans:next()):eq("x._y;\n")
        end)

        test("x:keyword", function()
          parser:parse("x:default")
          assert(trans:next()):eq("x._default;\n")
        end)

        test("x:y=z", function()
          parser:parse("x:y=z")
          assert(trans:next()):eq("(x._y=z);\n")
        end)

        test("x:keyword=y", function()
          parser:parse("x:default=y")
          assert(trans:next()):eq("(x._default=y);\n")
        end)
      end)

      suite("!", function()
        test("super!method()", function()
          parser:parse("super!m()")
          assert(trans:next()):eq('dogma.super(this, "_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_m")();\n')
        end)

        test("x!y", function()
          parser:parse("x!y")
          assert(trans:next()):eq("x._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_y;\n")
        end)

        test("x!keyword", function()
          parser:parse("x!default")
          assert(trans:next()):eq("x._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_default;\n")
        end)

        test("x!y=z", function()
          parser:parse("x!y=z")
          assert(trans:next()):eq("(x._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_y=z);\n")
        end)

        test("x!keyword=y", function()
          parser:parse("x!default=y")
          assert(trans:next()):eq("(x._pvt_e6eace38db3ac6dfd7468b60ae7a70ae_default=y);\n")
        end)
      end)

      test("?[]", function()
        parser:parse("x?[y]")
        assert(trans:next()):eq("dogma.getOptionalItem(x, y);\n")
      end)

      suite("[]", function()
        test("Exp [ Name ]", function()
          parser:parse("x[y]")
          assert(trans:next()):eq("dogma.getItem(x, y);\n")
        end)

        test("Exp [ Name ] = Exp", function()
          parser:parse("x[y] = z")
          assert(trans:next()):eq('dogma.setItem("=", x, y, z);\n')
        end)

        test("Exp [ num ]", function()
          parser:parse("x[123]")
          assert(trans:next()):eq("dogma.getItem(x, 123);\n")
        end)

        test("Exp [ text ]", function()
          parser:parse('x["id"]')
          assert(trans:next()):eq('dogma.getItem(x, "id");\n')
        end)
      end)

      suite("is", function()
        test("a is b", function()
          parser:parse("x is text")
          assert(trans:next()):eq("dogma.is(x, text);\n")
        end)

        test("a is [b, c]", function()
          parser:parse("x is [text, num]")
          assert(trans:next()):eq("dogma.is(x, [text, num]);\n")
        end)

        test("is not", function()
          parser:parse("x is not text")
          assert(trans:next()):eq("dogma.isNot(x, text);\n")
        end)
      end)

      test("a as b", function()
        parser:parse("x as y")
        assert(trans:next()):eq("dogma.setTypeToObject(x, y);\n")
      end)

      suite("in", function()
        test("in", function()
          parser:parse("x in array")
          assert(trans:next()):eq("dogma.includes(array, x);\n")
        end)

        test("not in", function()
          parser:parse("x not in array")
          assert(trans:next()):eq("!dogma.includes(array, x);\n")
        end)
      end)

      test("like", function()
        parser:parse("x like y")
        assert(trans:next()):eq("dogma.like(x, y);\n")
      end)

      test("notLike", function()
        parser:parse("x not like y")
        assert(trans:next()):eq("dogma.notLike(x, y);\n")
      end)
    end):tags("binop")

    ----------------
    -- ternary op --
    ----------------
    suite("ternary op", function()
      test("x[y, z]", function()
        parser:parse("x[y, z]")
        assert(trans:next()):eq("dogma.getSlice(x, y, z);\n")
      end)

      test("o.m[x, y]", function()
        parser:parse("o.m()[123, 456]")
        assert(trans:next()):eq("dogma.getSlice(o.m(), 123, 456);\n")
      end)
    end):tags("ternaryop")

    ----------
    -- call --
    ----------
    suite("call", function()
      test("x()", function()
        parser:parse("x()")
        assert(trans:next()):eq("x();\n")
      end)

      test("x(1)", function()
        parser:parse("x(1)")
        assert(trans:next()):eq("x(1);\n")
      end)

      test("x(1, 2)", function()
        parser:parse("x(1, 2)")
        assert(trans:next()):eq("x(1, 2);\n")
      end)

      test("x(a=1)", function()
        parser:parse("x(a=1)")
        assert(trans:next()):eq("x({'a': 1});\n")
      end)

      test("x((a=1))", function()
        parser:parse("x((a=1))")
        assert(trans:next()):eq("x((a=1));\n")
      end)

      test("x(a=1, b=2)", function()
        parser:parse("x(a=1, b=2)")
        assert(trans:next()):eq("x({'a': 1, 'b': 2});\n")
      end)

      test("x(12, 34, a=1, b=2, c, d=4)", function()
        parser:parse("x(12, 34, a=1, b=2, c, d=4)")
        assert(trans:next()):eq("x(12, 34, {'a': 1, 'b': 2, 'c': c, 'd': 4});\n")
      end)
    end)
  end)

  ----------
  -- pack --
  ----------
  suite("pack", function()
    suite("list", function()
      test("pack[*]", function()
        parser:parse("x[*]")
        assert(trans:next()):eq("dogma.cloneL(x);\n")
      end)

      test("pack[*,item,item]", function()
        parser:parse("x[*,1+2,3+4]")
        assert(trans:next()):eq("dogma.cloneL(x, (1+2), (3+4));\n")
      end)

      test("pack[*,item,...items,item]", function()
        parser:parse("x[*,1+2,...y,3+4]")
        assert(trans:next()):eq("dogma.cloneL(x, (1+2), ...(y), (3+4));\n")
      end)
    end)

    suite("map", function()
      test("pack{*}", function()
        parser:parse("x{*}")
        assert(trans:next()):eq("dogma.clone(x, {}, {}, [], []);\n")
      end)

      test("pack{*[*],field}", function(params)
        parser:parse(string.format("x{%s,field}", params[1]))
        assert(trans:next()):eq(string.format('dogma.%s(x, {"field": field}, {}, [], []);\n', params[2]))
      end):iter(
        {subtitle = " # *", params = {"*", "clone"}},
        {subtitle = " # **", params = {"**", "copyAndUpdate"}}
      )

      test("pack{*[*],-field}", function(params)
        parser:parse(string.format("x{%s,-field}", params[1]))
        assert(trans:next()):eq(string.format('dogma.%s(x, {}, {}, ["field"], []);\n', params[2]))
      end):iter(
        {subtitle = " # *", params = {"*", "clone"}},
        {subtitle = " # **", params = {"**", "copyAndUpdate"}}
      )

      test("pack{*[*],f1=val1}", function(params)
        parser:parse(string.format("x{%s,f1=1+2}", params[1]))
        assert(trans:next()):eq(string.format('dogma.%s(x, {"f1": (1+2)}, {}, [], []);\n', params[2]))
      end):iter(
        {subtitle = " # *", params = {"*", "clone"}},
        {subtitle = " # **", params = {"**", "copyAndUpdate"}}
      )

      test("pack{*[*],f1=val1,f2=val2}", function(params)
        parser:parse(string.format("x{%s,f1=val1,f2=val2}", params[1]))
        assert(trans:next()):eq(string.format('dogma.%s(x, {"f1": val1, "f2": val2}, {}, [], []);\n', params[2]))
      end):iter(
        {subtitle = " # *", params = {"*", "clone"}},
        {subtitle = " # **", params = {"**", "copyAndUpdate"}}
      )

      test("pack{*[*],field=>field}", function(params)
        parser:parse(string.format("x{%s,db=>database,user=>username}", params[1]))
        assert(trans:next()):eq(string.format('dogma.%s(x, {}, {"database": "db", "username": "user"}, [], []);\n', params[2]))
      end):iter(
        {subtitle = " # *", params = {"*", "clone"}},
        {subtitle = " # **", params = {"**", "copyAndUpdate"}}
      )

      test("pack{*[*],...obj}", function(params)
        parser:parse(string.format("x{%s,...obj}", params[1]))
        assert(trans:next()):eq(string.format('dogma.%s(x, {}, {}, [], [obj]);\n', params[2]))
      end):iter(
        {subtitle = " # *", params = {"*", "clone"}},
        {subtitle = " # **", params = {"**", "copyAndUpdate"}}
      )

      test("pack{*[*],f1=val1,-f2,f3=val3,-f4,...obj}", function(params)
        parser:parse(string.format("x{%s,f1=val1,-f2,f3=val3,-f4,...obj}", params[1]))
        assert(trans:next()):eq(string.format('dogma.%s(x, {"f1": val1, "f3": val3}, {}, ["f2", "f4"], [obj]);\n', params[2]))
      end):iter(
        {subtitle = " # *", params = {"*", "clone"}},
        {subtitle = " # **", params = {"**", "copyAndUpdate"}}
      )

      test("pack{}", function()
        parser:parse("x{}")
        assert(trans:next()):eq("dogma.pack(x);\n")
      end)

      test("pack{name}", function()
        parser:parse("x{name}")
        assert(trans:next()):eq('dogma.pack(x, "name");\n')
      end)

      test("pack{name, name=value}", function()
        parser:parse("x{f1,f2=val}")
        assert(trans:next()):eq('dogma.pack(x, "f1", {name: "f2", value: val});\n')
      end)

      test("pack{name,...obj}", function()
        parser:parse("x{f1,f2=val,...obj}")
        assert(trans:next()):eq('dogma.pack(x, "f1", {name: "f2", value: val}, {value: obj});\n')
      end)

      test("pack{name, name=value, name=value}", function()
        parser:parse("x{f1,f2=123,f3=456}")
        assert(trans:next()):eq('dogma.pack(x, "f1", {name: "f2", value: 123}, {name: "f3", value: 456});\n')
      end)

      test("pack{name, .name, :name}", function()
        parser:parse("x{a,.b,:c,!d}")
        assert(trans:next()):eq('dogma.pack(x, "a", "b", "_c", "_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_d");\n')
      end)
    end)
  end):tags("pack")

  ------------
  -- update --
  ------------
  suite("update", function()
    test("update.{a}", function()
      parser:parse("x.{a}")
      assert(trans:next()):eq('dogma.update(x, {name: "a", visib: ".", assign: "=", value: a});\n')
    end)

    test("update:{a}", function()
      parser:parse("x:{a}")
      assert(trans:next()):eq('dogma.update(x, {name: "a", visib: ":", assign: "=", value: a});\n')
    end)

    test("update!{a}", function()
      parser:parse("x!{a}")
      assert(trans:next()):eq('dogma.update(x, {name: "_pvt_e6eace38db3ac6dfd7468b60ae7a70ae_a", visib: "!", assign: "=", value: a});\n')
    end)

    test("update:{{a} = exp}", function()
      parser:parse("x:{{a} = 1+2}")
      assert(trans:next()):eq('dogma.update(x, {name: ["a"], visib: ":", assign: "=", value: (1+2), type: "mapped"});\n')
    end)

    test("update:{{a, b} = exp}", function()
      parser:parse("x:{{a, b} = 1+2}")
      assert(trans:next()):eq('dogma.update(x, {name: ["a", "b"], visib: ":", assign: "=", value: (1+2), type: "mapped"});\n')
    end)

    test("update:{(a, b) = exp}", function()
      parser:parse("x:{(a, b) = 1+2}")
      assert(trans:next()):eq('dogma.update(x, {name: ["a", "b"], visib: ":", assign: "=", value: (1+2), type: "extended"});\n')
    end)

    test("update.{a=exp, b::=exp, c?=exp, d.=exp}", function()
      parser:parse("x.{a=1+2, b::=3+4, c?=5+6, d.=7+8}")
      assert(trans:next()):eq('dogma.update(x, {name: "a", visib: ".", assign: "=", value: (1+2)}, {name: "b", visib: ".", assign: "::=", value: (3+4)}, {name: "c", visib: ".", assign: "?=", value: (5+6)}, {name: "d", visib: ".", assign: ".=", value: (7+8)});\n')
    end)
  end):tags("update")

  ----------
  -- misc --
  ----------
  suite("misc", function()
    test("terminal terminal - error", function()
      parser:parse("x y")
      assert(function() parser:next() end):raises("anonymous code (1, 3): terminal can't follow to other terminal.")
    end)
  end)
end):tags("exp")
